@@ -4,9 +4,8 @@
 // LICENSE file in the root directory of this source tree.
 
 use super::{constraints::CompositionPoly, StarkDomain, TracePolyTable};
-use air::{Air, DeepCompositionCoefficients, EvaluationFrame};
-use core::marker::PhantomData;
-use math::{add_in_place, fft, log2, mul_acc, polynom, FieldElement, StarkField};
+use air::{Air, DeepCompositionCoefficients};
+use math::{add_in_place, fft, log2, mul_acc, polynom, ExtensionOf, FieldElement, StarkField};
 use utils::{collections::Vec, iter_mut};
 
 #[cfg(feature = "concurrent")]
@@ -14,27 +13,28 @@ use utils::iterators::*;
 
 // DEEP COMPOSITION POLYNOMIAL
 // ================================================================================================
-pub struct DeepCompositionPoly<A: Air, E: FieldElement<BaseField = A::BaseField>> {
+pub struct DeepCompositionPoly<E: FieldElement> {
     coefficients: Vec<E>,
     cc: DeepCompositionCoefficients<E>,
     z: E,
     field_extension: bool,
-    _air: PhantomData<A>,
 }
 
-impl<A: Air, E: FieldElement<BaseField = A::BaseField>> DeepCompositionPoly<A, E> {
+impl<E: FieldElement> DeepCompositionPoly<E> {
     // CONSTRUCTOR
     // --------------------------------------------------------------------------------------------
     /// Returns a new DEEP composition polynomial. Initially, this polynomial will be empty, and
     /// the intent is to populate the coefficients via add_trace_polys() and add_constraint_polys()
     /// methods.
-    pub fn new(air: &A, z: E, cc: DeepCompositionCoefficients<E>) -> Self {
+    pub fn new<A>(air: &A, z: E, cc: DeepCompositionCoefficients<E>) -> Self
+    where
+        A: Air<BaseField = E::BaseField>,
+    {
         DeepCompositionPoly {
             coefficients: vec![],
             cc,
             z,
             field_extension: !air.options().field_extension().is_none(),
-            _air: PhantomData,
         }
     }
 
@@ -58,7 +58,7 @@ impl<A: Air, E: FieldElement<BaseField = A::BaseField>> DeepCompositionPoly<A, E
     ///
     /// - Compute polynomials T'_i(x) = (T_i(x) - T_i(z)) / (x - z) and
     ///   T''_i(x) = (T_i(x) - T_i(z * g)) / (x - z * g) for all i, where T_i(x) is a trace
-    ///   polynomial for register i.
+    ///   polynomial for column i.
     /// - Then, combine together all T'_i(x) polynomials using random liner combination as
     ///   T(x) = sum(T'_i(x) * cc'_i + T''_i(x) * cc''_i) for all i, where cc'_i and cc''_i are
     ///   the coefficients for the random linear combination drawn from the public coin.
@@ -70,24 +70,19 @@ impl<A: Air, E: FieldElement<BaseField = A::BaseField>> DeepCompositionPoly<A, E
     /// Note that evaluations of T_i(z) and T_i(z * g) are passed in via the `ood_frame` parameter.
     pub fn add_trace_polys(
         &mut self,
-        trace_polys: TracePolyTable<A::BaseField>,
-        ood_frame: EvaluationFrame<E>,
+        trace_polys: TracePolyTable<E>,
+        ood_trace_states: Vec<Vec<E>>,
     ) {
         assert!(self.coefficients.is_empty());
 
         // compute a second out-of-domain point offset from z by exactly trace generator; this
         // point defines the "next" computation state in relation to point z
         let trace_length = trace_polys.poly_size();
-        let g = E::from(A::BaseField::get_root_of_unity(log2(trace_length)));
+        let g = E::from(E::BaseField::get_root_of_unity(log2(trace_length)));
         let next_z = self.z * g;
 
-        // cache state of registers at points z and z * g
-        let trace_state1 = ood_frame.current();
-        let trace_state2 = ood_frame.next();
-
         // combine trace polynomials into 2 composition polynomials T'(x) and T''(x), and if
         // we are using a field extension, also T'''(x)
-        let polys = trace_polys.into_vec();
         let mut t1_composition = E::zeroed_vector(trace_length);
         let mut t2_composition = E::zeroed_vector(trace_length);
         let mut t3_composition = if self.field_extension {
@@ -95,35 +90,69 @@ impl<A: Air, E: FieldElement<BaseField = A::BaseField>> DeepCompositionPoly<A, E
         } else {
             Vec::new()
         };
-        for (i, poly) in polys.into_iter().enumerate() {
+
+        // index of a trace polynomial; we declare it here so that we can maintain index continuity
+        // across all trace segments
+        let mut i = 0;
+
+        // --- merge polynomials of the main trace segment ----------------------------------------
+        for poly in trace_polys.main_trace_polys() {
             // compute T'(x) = T(x) - T(z), multiply it by a pseudo-random coefficient,
             // and add the result into composition polynomial
-            acc_poly(
+            acc_trace_poly::<E::BaseField, E>(
                 &mut t1_composition,
-                &poly,
-                trace_state1[i],
+                poly,
+                ood_trace_states[0][i],
                 self.cc.trace[i].0,
             );
 
             // compute T''(x) = T(x) - T(z * g), multiply it by a pseudo-random coefficient,
             // and add the result into composition polynomial
-            acc_poly(
+            acc_trace_poly::<E::BaseField, E>(
                 &mut t2_composition,
-                &poly,
-                trace_state2[i],
+                poly,
+                ood_trace_states[1][i],
                 self.cc.trace[i].1,
             );
 
             // when extension field is enabled, compute T'''(x) = T(x) - T(z_conjugate), multiply
             // it by a pseudo-random coefficient, and add the result into composition polynomial
             if self.field_extension {
-                acc_poly(
+                acc_trace_poly::<E::BaseField, E>(
                     &mut t3_composition,
-                    &poly,
-                    trace_state1[i].conjugate(),
+                    poly,
+                    ood_trace_states[0][i].conjugate(),
                     self.cc.trace[i].2,
                 );
             }
+
+            i += 1;
+        }
+
+        // --- merge polynomials of the auxiliary trace segments ----------------------------------
+
+        // since trace polynomials are already in an extension field (when extension fields are
+        // used), we don't apply conjugate composition to them
+        for poly in trace_polys.aux_trace_polys() {
+            // compute T'(x) = T(x) - T(z), multiply it by a pseudo-random coefficient,
+            // and add the result into composition polynomial
+            acc_trace_poly::<E, E>(
+                &mut t1_composition,
+                poly,
+                ood_trace_states[0][i],
+                self.cc.trace[i].0,
+            );
+
+            // compute T''(x) = T(x) - T(z * g), multiply it by a pseudo-random coefficient,
+            // and add the result into composition polynomial
+            acc_trace_poly::<E, E>(
+                &mut t2_composition,
+                poly,
+                ood_trace_states[1][i],
+                self.cc.trace[i].1,
+            );
+
+            i += 1;
         }
 
         // divide the composition polynomials by (x - z), (x - z * g), and (x - z_conjugate)
@@ -154,7 +183,7 @@ impl<A: Air, E: FieldElement<BaseField = A::BaseField>> DeepCompositionPoly<A, E
     /// Note that evaluations of H_i(x) at z^m are passed in via the `ood_evaluations` parameter.
     pub fn add_composition_poly(
         &mut self,
-        composition_poly: CompositionPoly<A::BaseField, E>,
+        composition_poly: CompositionPoly<E>,
         ood_evaluations: Vec<E>,
     ) {
         assert!(!self.coefficients.is_empty());
@@ -176,7 +205,7 @@ impl<A: Air, E: FieldElement<BaseField = A::BaseField>> DeepCompositionPoly<A, E
 
         // add H'_i(x) * cc_i for all i into the DEEP composition polynomial
         for (i, poly) in column_polys.into_iter().enumerate() {
-            mul_acc(&mut self.coefficients, &poly, self.cc.constraints[i]);
+            mul_acc::<E, E>(&mut self.coefficients, &poly, self.cc.constraints[i]);
         }
         assert_eq!(self.poly_size() - 2, self.degree());
     }
@@ -196,9 +225,9 @@ impl<A: Air, E: FieldElement<BaseField = A::BaseField>> DeepCompositionPoly<A, E
         let mut result = E::zeroed_vector(self.coefficients.len());
 
         // this is equivalent to C(x) * cc_0
-        mul_acc(&mut result, &self.coefficients, self.cc.degree.0);
+        mul_acc::<E, E>(&mut result, &self.coefficients, self.cc.degree.0);
         // this is equivalent to C(x) * x * cc_1
-        mul_acc(
+        mul_acc::<E, E>(
             &mut result[1..],
             &self.coefficients[..(self.coefficients.len() - 1)],
             self.cc.degree.1,
@@ -211,7 +240,7 @@ impl<A: Air, E: FieldElement<BaseField = A::BaseField>> DeepCompositionPoly<A, E
     // LOW-DEGREE EXTENSION
     // --------------------------------------------------------------------------------------------
     /// Evaluates DEEP composition polynomial over the specified LDE domain and returns the result.
-    pub fn evaluate(self, domain: &StarkDomain<A::BaseField>) -> Vec<E> {
+    pub fn evaluate(self, domain: &StarkDomain<E::BaseField>) -> Vec<E> {
         fft::evaluate_poly_with_offset(
             &self.coefficients,
             domain.trace_twiddles(),
@@ -247,11 +276,11 @@ fn merge_trace_compositions<E: FieldElement>(mut polys: Vec<Vec<E>>, divisors: V
     result
 }
 
-/// Computes (P(x) - value) * k and saves the result into the accumulator
-fn acc_poly<B, E>(accumulator: &mut Vec<E>, poly: &[B], value: E, k: E)
+/// Computes (P(x) - value) * k and saves the result into the accumulator.
+fn acc_trace_poly<F, E>(accumulator: &mut [E], poly: &[F], value: E, k: E)
 where
-    B: StarkField,
-    E: FieldElement<BaseField = B>,
+    F: FieldElement,
+    E: FieldElement<BaseField = F::BaseField> + ExtensionOf<F>,
 {
     mul_acc(accumulator, poly, k);
     let adjusted_tz = value * k;
