@@ -3,10 +3,10 @@
 // This source code is licensed under the MIT license found in the
 // LICENSE file in the root directory of this source tree.
 
-use super::{trace::TraceLde, ProverError, StarkDomain};
+use super::{ConstraintDivisor, Matrix, ProverError, StarkDomain};
 
 mod boundary;
-use boundary::BoundaryConstraintGroup;
+use boundary::BoundaryConstraints;
 
 mod periodic_table;
 use periodic_table::PeriodicValueTable;
