@@ -44,13 +44,14 @@
 extern crate alloc;
 
 pub use air::{
-    proof::StarkProof, Air, AirContext, Assertion, BoundaryConstraint, BoundaryConstraintGroup,
-    ConstraintCompositionCoefficients, ConstraintDivisor, DeepCompositionCoefficients,
-    EvaluationFrame, FieldExtension, HashFunction, ProofOptions, TraceInfo,
-    TransitionConstraintDegree, TransitionConstraintGroup,
+    proof::StarkProof, Air, AirContext, Assertion, AuxTraceRandElements, BoundaryConstraint,
+    BoundaryConstraintGroup, ConstraintCompositionCoefficients, ConstraintDivisor,
+    DeepCompositionCoefficients, EvaluationFrame, FieldExtension, HashFunction, ProofOptions,
+    TraceInfo, TraceLayout, TransitionConstraintDegree, TransitionConstraintGroup,
 };
 pub use utils::{
     iterators, ByteReader, ByteWriter, Deserializable, DeserializationError, Serializable,
+    SliceReader,
 };
 
 use fri::FriProver;
@@ -66,7 +67,7 @@ use math::{
 pub use crypto;
 use crypto::{
     hashers::{Blake3_192, Blake3_256, Sha3_256},
-    ElementHasher,
+    ElementHasher, MerkleTree,
 };
 
 #[cfg(feature = "std")]
@@ -77,17 +78,20 @@ use math::log2;
 use std::time::Instant;
 
 mod domain;
-use domain::StarkDomain;
+pub use domain::StarkDomain;
+
+mod matrix;
+pub use matrix::Matrix;
 
 mod constraints;
-use constraints::{ConstraintCommitment, ConstraintEvaluator};
+use constraints::{CompositionPoly, ConstraintCommitment, ConstraintEvaluator};
 
 mod composer;
 use composer::DeepCompositionPoly;
 
 mod trace;
-use trace::TracePolyTable;
 pub use trace::{Trace, TraceTable, TraceTableFragment};
+use trace::{TraceCommitment, TraceLde, TracePolyTable};
 
 mod channel;
 use channel::ProverChannel;
@@ -192,7 +196,7 @@ pub trait Prover {
     /// execution `trace` is valid against this prover's AIR.
     /// TODO: make this function un-callable externally?
     #[doc(hidden)]
-    fn generate_proof<E, H>(&self, trace: Self::Trace) -> Result<StarkProof, ProverError>
+    fn generate_proof<E, H>(&self, mut trace: Self::Trace) -> Result<StarkProof, ProverError>
     where
         E: FieldElement<BaseField = Self::BaseField>,
         H: ElementHasher<BaseField = Self::BaseField>,
@@ -209,18 +213,12 @@ pub trait Prover {
         // execution of the computation for the provided public inputs.
         let air = Self::Air::new(trace.get_info(), pub_inputs, self.options().clone());
 
-        // make sure the specified trace is valid against the AIR. This checks validity of both,
-        // assertions and state transitions. we do this in debug mode only because this is a very
-        // expensive operation.
-        #[cfg(debug_assertions)]
-        trace.validate(&air);
-
         // create a channel which is used to simulate interaction between the prover and the
         // verifier; the channel will be used to commit to values and to draw randomness that
         // should come from the verifier.
         let mut channel = ProverChannel::<Self::Air, E, H>::new(&air, pub_inputs_bytes);
 
-        // 1 ----- extend execution trace ---------------------------------------------------------
+        // 1 ----- Commit to the execution trace --------------------------------------------------
 
         // build computation domain; this is used later for polynomial evaluations
         #[cfg(feature = "std")]
@@ -233,33 +231,59 @@ pub trait Prover {
             now.elapsed().as_millis()
         );
 
-        // extend the execution trace; this interpolates each register of the trace into a
-        // polynomial, and then evaluates the polynomial over the LDE domain; each of the trace
-        // polynomials has degree = trace_length - 1
-        let (extended_trace, trace_polys) = trace.extend(&domain);
-        #[cfg(feature = "std")]
-        debug!(
-        "Extended execution trace of {} registers from 2^{} to 2^{} steps ({}x blowup) in {} ms",
-        extended_trace.width(),
-        log2(trace_polys.poly_size()),
-        log2(extended_trace.len()),
-        extended_trace.blowup(),
-        now.elapsed().as_millis()
-    );
+        // extend the main execution trace and build a Merkle tree from the extended trace
+        let (main_trace_lde, main_trace_tree, main_trace_polys) =
+            self.build_trace_commitment::<Self::BaseField, H>(trace.main_segment(), &domain);
 
-        // 2 ----- commit to the extended execution trace -----------------------------------------
-        #[cfg(feature = "std")]
-        let now = Instant::now();
-        let trace_tree = extended_trace.build_commitment::<H>();
-        channel.commit_trace(*trace_tree.root());
-        #[cfg(feature = "std")]
-        debug!(
-            "Committed to extended execution trace by building a Merkle tree of depth {} in {} ms",
-            trace_tree.depth(),
-            now.elapsed().as_millis()
+        // commit to the LDE of the main trace by writing the root of its Merkle tree into
+        // the channel
+        channel.commit_trace(*main_trace_tree.root());
+
+        // initialize trace commitment and trace polynomial table structs with the main trace
+        // data; for multi-segment traces these structs will be used as accumulators of all
+        // trace segments
+        let mut trace_commitment = TraceCommitment::new(
+            main_trace_lde,
+            main_trace_tree,
+            domain.trace_to_lde_blowup(),
         );
+        let mut trace_polys = TracePolyTable::new(main_trace_polys);
 
-        // 3 ----- evaluate constraints -----------------------------------------------------------
+        // build auxiliary trace segments (if any), and append the resulting segments to trace
+        // commitment and trace polynomial table structs
+        let mut aux_trace_segments = Vec::new();
+        let mut aux_trace_rand_elements = AuxTraceRandElements::new();
+        for i in 0..trace.layout().num_aux_segments() {
+            // draw a set of random elements required to build an auxiliary trace segment
+            let rand_elements = channel.get_aux_trace_segment_rand_elements(i);
+
+            // build the trace segment
+            let aux_segment = trace
+                .build_aux_segment(&aux_trace_segments, &rand_elements)
+                .expect("failed build auxiliary trace segment");
+
+            // extend the auxiliary trace segment and build a Merkle tree from the extended trace
+            let (aux_segment_lde, aux_segment_tree, aux_segment_polys) =
+                self.build_trace_commitment::<E, H>(&aux_segment, &domain);
+
+            // commit to the LDE of the extended auxiliary trace segment  by writing the root of
+            // its Merkle tree into the channel
+            channel.commit_trace(*aux_segment_tree.root());
+
+            // append the segment to the trace commitment and trace polynomial table structs
+            trace_commitment.add_segment(aux_segment_lde, aux_segment_tree);
+            trace_polys.add_aux_segment(aux_segment_polys);
+            aux_trace_rand_elements.add_segment_elements(rand_elements);
+            aux_trace_segments.push(aux_segment);
+        }
+
+        // make sure the specified trace (including auxiliary segments) is valid against the AIR.
+        // This checks validity of both, assertions and state transitions. We do this in debug
+        // mode only because this is a very expensive operation.
+        #[cfg(debug_assertions)]
+        trace.validate(&air, &aux_trace_segments, &aux_trace_rand_elements);
+
+        // 2 ----- evaluate constraints -----------------------------------------------------------
         // evaluate constraints specified by the AIR over the constraint evaluation domain, and
         // compute random linear combinations of these evaluations using coefficients drawn from
         // the channel; this step evaluates only constraint numerators, thus, only constraints with
@@ -269,8 +293,8 @@ pub trait Prover {
         #[cfg(feature = "std")]
         let now = Instant::now();
         let constraint_coeffs = channel.get_constraint_composition_coeffs();
-        let evaluator = ConstraintEvaluator::new(&air, constraint_coeffs);
-        let constraint_evaluations = evaluator.evaluate(&extended_trace, &domain);
+        let evaluator = ConstraintEvaluator::new(&air, aux_trace_rand_elements, constraint_coeffs);
+        let constraint_evaluations = evaluator.evaluate(trace_commitment.trace_table(), &domain);
         #[cfg(feature = "std")]
         debug!(
             "Evaluated constraints over domain of 2^{} elements in {} ms",
@@ -278,7 +302,7 @@ pub trait Prover {
             now.elapsed().as_millis()
         );
 
-        // 4 ----- commit to constraint evaluations -----------------------------------------------
+        // 3 ----- commit to constraint evaluations -----------------------------------------------
 
         // first, build constraint composition polynomial from the constraint evaluation table:
         // - divide all constraint evaluation columns by their respective divisors
@@ -291,36 +315,21 @@ pub trait Prover {
         let composition_poly = constraint_evaluations.into_poly()?;
         #[cfg(feature = "std")]
         debug!(
-        "Converted constraint evaluations into {} composition polynomial columns of degree {} in {} ms",
-        composition_poly.num_columns(),
-        composition_poly.column_degree(),
-        now.elapsed().as_millis()
-    );
-
-        // then, evaluate composition polynomial columns over the LDE domain
-        #[cfg(feature = "std")]
-        let now = Instant::now();
-        let composed_evaluations = composition_poly.evaluate(&domain);
-        #[cfg(feature = "std")]
-        debug!(
-            "Evaluated composition polynomial columns over LDE domain (2^{} elements) in {} ms",
-            log2(composed_evaluations[0].len()),
+            "Converted constraint evaluations into {} composition polynomial columns of degree {} in {} ms",
+            composition_poly.num_columns(),
+            composition_poly.column_degree(),
             now.elapsed().as_millis()
         );
 
-        // finally, commit to the composition polynomial evaluations
-        #[cfg(feature = "std")]
-        let now = Instant::now();
-        let constraint_commitment = ConstraintCommitment::<E, H>::new(composed_evaluations);
+        // then, build a commitment to the evaluations of the composition polynomial columns
+        let constraint_commitment =
+            self.build_constraint_commitment::<E, H>(&composition_poly, &domain);
+
+        // then, commit to the evaluations of constraints by writing the root of the constraint
+        // Merkle tree into the channel
         channel.commit_constraints(constraint_commitment.root());
-        #[cfg(feature = "std")]
-        debug!(
-            "Committed to composed evaluations by building a Merkle tree of depth {} in {} ms",
-            constraint_commitment.tree_depth(),
-            now.elapsed().as_millis()
-        );
 
-        // 5 ----- build DEEP composition polynomial ----------------------------------------------
+        // 4 ----- build DEEP composition polynomial ----------------------------------------------
         #[cfg(feature = "std")]
         let now = Instant::now();
 
@@ -336,8 +345,8 @@ pub trait Prover {
         // evaluate trace and constraint polynomials at the OOD point z, and send the results to
         // the verifier. the trace polynomials are actually evaluated over two points: z and z * g,
         // where g is the generator of the trace domain.
-        let ood_frame = trace_polys.get_ood_frame(z);
-        channel.send_ood_evaluation_frame(&ood_frame);
+        let ood_trace_states = trace_polys.get_ood_frame(z);
+        channel.send_ood_trace_states(&ood_trace_states);
 
         let ood_evaluations = composition_poly.evaluate_at(z);
         channel.send_ood_constraint_evaluations(&ood_evaluations);
@@ -349,7 +358,7 @@ pub trait Prover {
 
         // combine all trace polynomials together and merge them into the DEEP composition
         // polynomial
-        deep_composition_poly.add_trace_polys(trace_polys, ood_frame);
+        deep_composition_poly.add_trace_polys(trace_polys, ood_trace_states);
 
         // merge columns of constraint composition polynomial into the DEEP composition polynomial;
         deep_composition_poly.add_composition_poly(composition_poly, ood_evaluations);
@@ -369,7 +378,7 @@ pub trait Prover {
         // degree
         assert_eq!(domain.trace_length() - 1, deep_composition_poly.degree());
 
-        // 6 ----- evaluate DEEP composition polynomial over LDE domain ---------------------------
+        // 5 ----- evaluate DEEP composition polynomial over LDE domain ---------------------------
         #[cfg(feature = "std")]
         let now = Instant::now();
         let deep_evaluations = deep_composition_poly.evaluate(&domain);
@@ -386,7 +395,7 @@ pub trait Prover {
             now.elapsed().as_millis()
         );
 
-        // 7 ----- compute FRI layers for the composition polynomial ------------------------------
+        // 6 ----- compute FRI layers for the composition polynomial ------------------------------
         #[cfg(feature = "std")]
         let now = Instant::now();
         let mut fri_prover = FriProver::new(air.options().to_fri_options());
@@ -398,7 +407,7 @@ pub trait Prover {
             now.elapsed().as_millis()
         );
 
-        // 8 ----- determine query positions ------------------------------------------------------
+        // 7 ----- determine query positions ------------------------------------------------------
         #[cfg(feature = "std")]
         let now = Instant::now();
 
@@ -414,7 +423,7 @@ pub trait Prover {
             now.elapsed().as_millis()
         );
 
-        // 9 ----- build proof object -------------------------------------------------------------
+        // 8 ----- build proof object -------------------------------------------------------------
         #[cfg(feature = "std")]
         let now = Instant::now();
 
@@ -423,7 +432,7 @@ pub trait Prover {
 
         // query the execution trace at the selected position; for each query, we need the
         // state of the trace at that position + Merkle authentication path
-        let trace_queries = extended_trace.query(trace_tree, &query_positions);
+        let trace_queries = trace_commitment.query(&query_positions);
 
         // query the constraint commitment at the selected positions; for each query, we need just
         // a Merkle authentication path. this is because constraint evaluations for each step are
@@ -437,4 +446,94 @@ pub trait Prover {
 
         Ok(proof)
     }
+
+    /// Computes a low-degree extension (LDE) of the provided execution trace over the specified
+    /// domain and build a commitment to the extended trace.
+    ///
+    /// The extension is performed by interpolating each column of the execution trace into a
+    /// polynomial of degree = trace_length - 1, and then evaluating the polynomial over the LDE
+    /// domain.
+    ///
+    /// Trace commitment is computed by hashing each row of the extended execution trace, and then
+    /// building a Merkle tree from the resulting hashes.
+    fn build_trace_commitment<E, H>(
+        &self,
+        trace: &Matrix<E>,
+        domain: &StarkDomain<Self::BaseField>,
+    ) -> (Matrix<E>, MerkleTree<H>, Matrix<E>)
+    where
+        E: FieldElement<BaseField = Self::BaseField>,
+        H: ElementHasher<BaseField = Self::BaseField>,
+    {
+        // extend the execution trace
+        #[cfg(feature = "std")]
+        let now = Instant::now();
+        let trace_polys = trace.interpolate_columns();
+        let trace_lde = trace_polys.evaluate_columns_over(domain);
+        #[cfg(feature = "std")]
+        debug!(
+            "Extended execution trace of {} columns from 2^{} to 2^{} steps ({}x blowup) in {} ms",
+            trace_lde.num_cols(),
+            log2(trace_polys.num_rows()),
+            log2(trace_lde.num_rows()),
+            domain.trace_to_lde_blowup(),
+            now.elapsed().as_millis()
+        );
+
+        // build trace commitment
+        #[cfg(feature = "std")]
+        let now = Instant::now();
+        let trace_tree = trace_lde.commit_to_rows();
+        #[cfg(feature = "std")]
+        debug!(
+            "Computed execution trace commitment (Merkle tree of depth {}) in {} ms",
+            trace_tree.depth(),
+            now.elapsed().as_millis()
+        );
+
+        (trace_lde, trace_tree, trace_polys)
+    }
+
+    /// Evaluates constraint composition polynomial over the LDE domain and builds a commitment
+    /// to these evaluations.
+    ///
+    /// The evaluation is done by evaluating each composition polynomial column over the LDE
+    /// domain.
+    ///
+    /// The commitment is computed by hashing each row in the evaluation matrix, and then building
+    /// a Merkle tree from the resulting hashes.
+    fn build_constraint_commitment<E, H>(
+        &self,
+        composition_poly: &CompositionPoly<E>,
+        domain: &StarkDomain<Self::BaseField>,
+    ) -> ConstraintCommitment<E, H>
+    where
+        E: FieldElement<BaseField = Self::BaseField>,
+        H: ElementHasher<BaseField = Self::BaseField>,
+    {
+        // evaluate composition polynomial columns over the LDE domain
+        #[cfg(feature = "std")]
+        let now = Instant::now();
+        let composed_evaluations = composition_poly.evaluate(domain);
+        #[cfg(feature = "std")]
+        debug!(
+            "Evaluated {} composition polynomial columns over LDE domain (2^{} elements) in {} ms",
+            composed_evaluations.num_cols(),
+            log2(composed_evaluations.num_rows()),
+            now.elapsed().as_millis()
+        );
+
+        // build constraint evaluation commitment
+        #[cfg(feature = "std")]
+        let now = Instant::now();
+        let commitment = composed_evaluations.commit_to_rows();
+        let constraint_commitment = ConstraintCommitment::new(composed_evaluations, commitment);
+        #[cfg(feature = "std")]
+        debug!(
+            "Computed constraint evaluation commitment (Merkle tree of depth {}) in {} ms",
+            constraint_commitment.tree_depth(),
+            now.elapsed().as_millis()
+        );
+        constraint_commitment
+    }
 }
